/*******************************************************************
 *
 * NetSaint Statistics Service for NT
 * Copyright (c) 2000 Ethan Galstad (netsaint@linuxbox.com)
 * Last Modified: 02-12-2000
 * License: GPL
 *
 * Credits: Most of the perfdata code is based on the code found
 *          int CDC's BackOrifice 2000.  See http://www.bo2k.com
 *			for more information.
 *
 * Description:
 * This service allows clients to check:
 *   1) physical and paged memory information on the local machine
 *   2) disk space either on a local drive or a share
 *   3) processor load (either average or for the first 4 individual
 *      processors)
 *
 *
 ******************************************************************/

#include <windows.h>
#include <winsock.h>
#include <winbase.h>
#include <stdlib.h>
#include <stdio.h>
#include <malloc.h>
#include <ctype.h>
#include <tchar.h>
#include <string.h>
#include <time.h>

#include "include\service.h"
#include "include\ntstat.h"

#include "include\perf.h"


int stop=0;
int server_port=DEFAULT_SERVER_PORT;

SOCKET listen_socket;

char authorized_hosts[MAX_AUTHORIZED_HOSTS][MAX_IPADDRESS_LENGTH];
char *config_file="\\Winnt\\system32\\ntstat.ini";

DOUBLE cpu_load_1_minute[4];
DOUBLE cpu_load_5_minute[4];
DOUBLE cpu_load_15_minute[4];

DOUBLE average_cpu_load_1_minute;
DOUBLE average_cpu_load_5_minute;
DOUBLE average_cpu_load_15_minute;


// start the service
VOID ServiceStart (DWORD dwArgc, LPTSTR *lpszArgv){
	int err;

	// tell the service manager that we're still starting up...
    if (!ReportStatusToSCMgr(SERVICE_START_PENDING,NO_ERROR,30000))
        goto cleanup;

	printf("starting service...\n");

	//read in the authorized hosts
	read_configuration();

	// tell the service manager that we're still starting up...
    if (!ReportStatusToSCMgr(SERVICE_START_PENDING,NO_ERROR,30000))
        goto cleanup;

	// start Winsock
	err=startWinsock();

	// stop the service if we couldn't initialize Winsock
	if(err)
			goto cleanup;

	// tell the service manager that we're still starting up...
    if (!ReportStatusToSCMgr(SERVICE_START_PENDING,NO_ERROR,30000))
        goto cleanup;

	// initialize server
	err=startServer();

	// if we cound't start the service, exit
	if(err)
			goto cleanup;

	// tell the service manager that we're up and running
    if (!ReportStatusToSCMgr(SERVICE_RUNNING,NO_ERROR,30000))
        goto cleanup;

	// keep handling connection requests until we should stop
	while(!stop){
		handle_connection_request();
		Sleep(500);		// wait 500ms, so we don't hog the CPU
		}

	// shutdown the service and cleanup
	cleanup:

	// tell the service manager that we're stopping...
	ReportStatusToSCMgr(SERVICE_STOP_PENDING,NO_ERROR,3000);

	// stop the server
	err=stopServer();

	// tell the service manager that we're stopping...
	ReportStatusToSCMgr(SERVICE_STOP_PENDING,NO_ERROR,3000);

	// cleanup Winsock
	err=stopWinsock();
	}



// stop the service
void ServiceStop(){

	// tell the service manager that we're stopping...
	ReportStatusToSCMgr(SERVICE_STOP_PENDING,NO_ERROR,3000);

	// set the stop flag
	stop=1;

	printf("Service stopping...\n");
	}



// read in configuration data
int read_configuration(void){
	FILE *fp;
	char input_buffer[MAX_INPUT_BUFFER];
	char *temp_ptr;
	int x;

	// clear the list of authorized hosts
	for(x=0;x<MAX_AUTHORIZED_HOSTS;x++)
		strcpy(authorized_hosts[x],"");

	fp=fopen("\\Winnt\\system32\\ntstat.ini","r");
	//fp=fopen("ntstat.ini","r");
	if(fp==NULL){
		printf("Error: Could not open config file '%s' for reading!\n",config_file);
		return 1;
		}

	x=0;
	while(fgets(input_buffer,sizeof(input_buffer)-1,fp)){

		strip(input_buffer);

		// skip comments
		if(input_buffer[0]=='#')
			continue;

		// skip blank lines
		if(input_buffer[0]==' ' || input_buffer[0]=='\r' || input_buffer[0]=='\n' || input_buffer[0]=='\x0')
				continue;

		printf("Buffer='%s'\n",input_buffer);

		// read in list of hosts that are authorized to connect
		if(strstr(input_buffer,"authorized_hosts=")==input_buffer){
			temp_ptr=strtok(input_buffer,"=");
			for(temp_ptr=strtok(NULL,",");temp_ptr!=NULL;temp_ptr=strtok(NULL,",")){
				if(x>=MAX_AUTHORIZED_HOSTS-1)
						break;
				strip(temp_ptr);
				printf("Address: '%s'\n",temp_ptr);
				strncpy(authorized_hosts[x],temp_ptr,MAX_IPADDRESS_LENGTH-1);
				authorized_hosts[x][MAX_IPADDRESS_LENGTH-1]='\x0';
				x++;
				}
			}

		// read in the port to use for communication
		else if(strstr(input_buffer,"server_port=")==input_buffer){
			temp_ptr=strtok(input_buffer,"=");
			temp_ptr=strtok(NULL,"\n");
			server_port=atoi(temp_ptr);
			}
		}

	fclose(fp);


	return 0;
	}



// initialize WinSock
int startWinsock(void){

	WORD wVersionRequested;
	WSADATA wsaData;
	int err;
 
	wVersionRequested=MAKEWORD(2,0);
 
	err=WSAStartup(wVersionRequested,&wsaData);
	if(err!=0){
		printf("Error: Could not initialize Winsock\n");
		return 1;
		}
 
	/* Confirm that the WinSock DLL supports 2.0.*/
	if(LOBYTE(wsaData.wVersion)!=2){
		printf("Error: Winsock 2.0 or greater is not installed\n");
	    WSACleanup();
		return 1; 
		}
 
	return 0;
	}



// cleanup WinSock
int stopWinsock(void){

	WSACleanup();

	return 0;
	}



// start the service
int startServer(void){
	struct in_addr bindAddr;
	SOCKADDR_IN saddr;
	BOOL sopt;
	unsigned long socketio;

	// create a socket
	listen_socket=socket(AF_INET,SOCK_STREAM,IPPROTO_TCP);

	if(listen_socket==INVALID_SOCKET){
		printf("Error: Could not create socket - WSAGetLastError returned %d\n",WSAGetLastError());
		return 1;
		}

	bindAddr.S_un.S_addr=INADDR_ANY;

	// set socket options
	sopt=TRUE;
	setsockopt(listen_socket,IPPROTO_TCP,TCP_NODELAY,(char *)&sopt,sizeof(BOOL));
	sopt=TRUE;
	setsockopt(listen_socket,SOL_SOCKET,SO_DONTLINGER,(char *)&sopt,sizeof(BOOL));

	// make the socket non-blocking
	socketio=1;
	ioctlsocket(listen_socket,FIONBIO,&socketio);

	// bind socket
	memset(&saddr,0,sizeof(SOCKADDR_IN));
	saddr.sin_addr=bindAddr;
	saddr.sin_family=AF_INET;
	saddr.sin_port=htons((WORD)server_port);
	
	if(bind(listen_socket,(SOCKADDR *)&saddr,sizeof(SOCKADDR_IN))==SOCKET_ERROR){
		printf("Error: Could not bind socket\n");
		closesocket(listen_socket);
		return 1;
		}

	// listen for connections
	if(listen(listen_socket,MAX_SERVER_CONNECTIONS)!=0){
		printf("Error: Could not listen on socket\n");
		closesocket(listen_socket);
		return 1;
		}

	// start monitoring CPU load
	CreateThread(NULL,0,(void *)&PerfData_GetProcessorLoad,NULL,0,NULL);
	
	return 0;
	}



// stop the service
int stopServer(void){

	// close the socket
	closesocket(listen_socket);

	return 0;
	}



// handles an incoming connection request
int handle_connection_request(void){
	SOCKET accept_socket;
	SOCKADDR_IN saddr;
	BOOL sopt;
	int len=sizeof(SOCKADDR_IN);
	int err;

	// try to accept an incoming connection
	accept_socket=accept(listen_socket,(SOCKADDR *)&saddr,&len);

	// there was no incoming connection request
	if(accept_socket==INVALID_SOCKET){
		printf("No connection request...\n");
		return 0;
		}

	printf("We received a connection request!\n");

	// set socket options
	sopt=TRUE;
	setsockopt(accept_socket,IPPROTO_TCP,TCP_NODELAY,(char *)&sopt,sizeof(BOOL));
	sopt=TRUE;
	setsockopt(accept_socket,SOL_SOCKET,SO_DONTLINGER,(char *)&sopt,sizeof(BOOL));

	// see if the connection is from an authorized host
	err=authorize_connection(saddr,len);
	if(err){
		printf("Host not authorized to connect!\n");
		closesocket(accept_socket);
		//ExitThread(1);
		return 1;
		}

	// handle the connection
	//CreateThread(NULL,0,(void *)&handle_connection,NULL,0,NULL);
	handle_connection(accept_socket);

	return 0;
	}



// make sure client is authorized to connect
int authorize_connection(SOCKADDR_IN saddr,int len){
	char connect_address[20];
	int x;

	sprintf(connect_address,"%u.%u.%u.%u",saddr.sin_addr.S_un.S_un_b.s_b1,saddr.sin_addr.S_un.S_un_b.s_b2,saddr.sin_addr.S_un.S_un_b.s_b3,saddr.sin_addr.S_un.S_un_b.s_b4);

	for(x=0;x<MAX_AUTHORIZED_HOSTS;x++){
		if(!strcmp(connect_address,authorized_hosts[x])){
			printf("Connection from %s accepted!\n",connect_address);
			return 0;
			}
		printf("authorized_hosts[%d]='%s'\n",x,authorized_hosts[x]);
		}

	printf("Connection from %s denied!\n",connect_address);

	return 1;
	}



// handles a client connection 
int handle_connection(SOCKET s){
	char send_buffer[MAX_INPUT_BUFFER];
	char recv_buffer[MAX_INPUT_BUFFER];
	int bytes_sent;
	int bytes_read;
	int err;

	// get data from the client 
	strcpy(recv_buffer,"");
	err=receive_data(s,recv_buffer,sizeof(recv_buffer),&bytes_read);
	if(err){
		printf("Receive error\n");
		closesocket(s);
		//ExitThread(0);
		return 0;
		}

	// process all client requests
	strcpy(send_buffer,"");
	process_requests(recv_buffer,sizeof(recv_buffer),send_buffer,sizeof(send_buffer));

	// send results back to client
	strcat(send_buffer,"\r\n");
	send_data(s,send_buffer,strlen(send_buffer),&bytes_sent);

	// wait a while for all data to get back to client
	Sleep(100);

	// close the connection	
	closesocket(s);

	//ExitThread(0);
	return 0;
	}



// sends data to client
int send_data(SOCKET s,char FAR *buffer,int buffer_length,int FAR *total_bytes_sent){
	int bytes_sent;

	bytes_sent=send(s,buffer,buffer_length,0);

	if(bytes_sent==SOCKET_ERROR){
		printf("Error with send()\n");
		return 1;
		}

	printf("Total bytes sent back to client=%d\n",bytes_sent);

	*total_bytes_sent=bytes_sent;

	return 0;
	}


// receives data from client
int receive_data(SOCKET s,char FAR *buffer,int max_length,int FAR *total_bytes_read){
	TIMEVAL tv;
	int bytes_read=0;
	int bytes_waiting;
	int nRet;
	fd_set rfds;
	char *newline;
	time_t connect_time;
	time_t current_time;

	time(&connect_time);

	// wait for all data up to a newline character or carriage return
	while(1){

		time(&current_time);
		if(current_time-connect_time>CONNECTION_TIMEOUT){
			printf("Connection timeout!\n");
			return 1;
			}

		// see if any data has arrived
		tv.tv_sec=CONNECTION_TIMEOUT;
		tv.tv_usec=0;
		FD_ZERO(&rfds);
		FD_SET(s,&rfds);
		nRet=select(1,&rfds,NULL,NULL,&tv);

		// error select()ing socket
		if(nRet==SOCKET_ERROR){
			printf("Socket error w/ select()\n");
			return 1;
			}

		// no data has arrived yet
		if(nRet==0){
			printf("No data has arrived!\n");
			//continue;
			return 1;
			}

		// error checking socket
		if(ioctlsocket(s,FIONREAD,&bytes_waiting)==SOCKET_ERROR){
			printf("Socket error w/ ioctlsocket()\n");
			return 1;
			}

		bytes_read+=recv(s,buffer+bytes_read,max_length-bytes_read,0);
		if(bytes_read==SOCKET_ERROR){
			printf("Socket error w/ recv()\n");
			return 1;
			}
		else{
			buffer[bytes_read]='\x0';
			printf("Total bytes so far: %d\n",bytes_read);
			}

		// break out if we've exceeded the buffer length
		if(bytes_read>=max_length){
			buffer[max_length-1]='\x0';
			printf("We exceeded the buffer length! %d>%d\n",bytes_read,max_length);
			break;
			}

		// break out it we've received a newline or carriage return character
		newline=strchr(buffer,(int)'\r');
		if(newline!=NULL){
			newline[0]='\x0';
			bytes_read=strlen(buffer);
			printf("We found a carriage return! Buffer='%s'\n",buffer);
			break;
			}
		newline=strchr(buffer,(int)'\n');
		if(newline!=NULL){
			newline[0]='\x0';
			bytes_read=strlen(buffer);
			printf("We found a newline character! Buffer='%s'\n",buffer);
			break;
			}
		}

	*total_bytes_read=bytes_read;
	
	return 0;
	}



// process all client requests
int process_requests(char FAR *recv_buffer,int recv_length,char FAR *send_buffer,int send_length){
	char temp_buffer[MAX_INPUT_BUFFER];
	char *temp_ptr=NULL;

	printf("process_requests() has recv buffer='%s'\n",recv_buffer);

	strcpy(temp_buffer,recv_buffer);

	// process each request
	for(temp_ptr=strtok(temp_buffer," ");temp_ptr!=NULL;temp_ptr=strtok(NULL," ")){
		strip(temp_ptr);
		process_request(temp_ptr,send_buffer+strlen(send_buffer),send_length-strlen(send_buffer));
		strncat(send_buffer," ",send_length-strlen(send_buffer));
		}

	printf("process_requests() has send buffer='%s'\n",send_buffer);

	return 0;
	}


// stips a buffer of white space, tabs, carriage returns, and newlines
int strip(char FAR *buffer){
	int lastchar;

	lastchar=strlen(buffer)-1;
	while(buffer[lastchar]==' ' || buffer[lastchar]=='\t' || buffer[lastchar]=='\r' || buffer[lastchar]=='\n'){
		buffer[lastchar]='\x0';
		lastchar--;
		}

	return 0;
	}



// process a single request
int process_request(char FAR *recv_buffer,char FAR *send_buffer,int send_length){
	char temp_buffer[MAX_INPUT_BUFFER];
	int temp_int=-1;
	int err;
	ULARGE_INTEGER temp_ularge;

	strcpy(temp_buffer,"");

	if(!strcmp(recv_buffer,"VER"))
		sprintf(temp_buffer,"%s",PROGRAM_VERSION);

	else if(!strcmp(recv_buffer,"TPHYSMEM")){
		get_total_physical_memory(&temp_int);
		sprintf(temp_buffer,"%d",temp_int);
		}

	else if(!strcmp(recv_buffer,"APHYSMEM")){
		get_available_physical_memory(&temp_int);
		sprintf(temp_buffer,"%d",temp_int);
		}

	else if(!strcmp(recv_buffer,"TPAGEMEM")){
		get_total_paged_memory(&temp_int);
		sprintf(temp_buffer,"%d",temp_int);
		}

	else if(!strcmp(recv_buffer,"APAGEMEM")){
		get_available_paged_memory(&temp_int);
		sprintf(temp_buffer,"%d",temp_int);
		}

	else if(strstr(recv_buffer,"TDISK")==recv_buffer){
		err=get_total_disk_space(recv_buffer+5,&temp_ularge);
		if(err)
			strcpy(temp_buffer,"-1");
		else
			sprintf(temp_buffer,"%I64d",temp_ularge);
		}

	else if(strstr(recv_buffer,"ADISK")==recv_buffer){
		err=get_available_disk_space(recv_buffer+5,&temp_ularge);
		if(err)
			strcpy(temp_buffer,"-1");
		else
			sprintf(temp_buffer,"%I64d",temp_ularge);
		}
	else if(!strcmp(recv_buffer,"CPULOAD1"))
		sprintf(temp_buffer,"%3.2lf",average_cpu_load_1_minute);
	else if(!strcmp(recv_buffer,"CPULOAD5"))
		sprintf(temp_buffer,"%3.2lf",average_cpu_load_5_minute);
	else if(!strcmp(recv_buffer,"CPULOAD15"))
		sprintf(temp_buffer,"%3.2lf",average_cpu_load_15_minute);

	else if(!strcmp(recv_buffer,"CPU0LOAD1"))
		sprintf(temp_buffer,"%3.2lf",cpu_load_1_minute[0]);
	else if(!strcmp(recv_buffer,"CPU0LOAD5"))
		sprintf(temp_buffer,"%3.2lf",cpu_load_5_minute[0]);
	else if(!strcmp(recv_buffer,"CPU0LOAD15"))
		sprintf(temp_buffer,"%3.2lf",cpu_load_15_minute[0]);

	else if(!strcmp(recv_buffer,"CPU1LOAD1"))
		sprintf(temp_buffer,"%3.2lf",cpu_load_1_minute[1]);
	else if(!strcmp(recv_buffer,"CPU1LOAD5"))
		sprintf(temp_buffer,"%3.2lf",cpu_load_5_minute[1]);
	else if(!strcmp(recv_buffer,"CPU1LOAD15"))
		sprintf(temp_buffer,"%3.2lf",cpu_load_15_minute[1]);

	else if(!strcmp(recv_buffer,"CPU2LOAD1"))
		sprintf(temp_buffer,"%3.2lf",cpu_load_1_minute[2]);
	else if(!strcmp(recv_buffer,"CPU2LOAD5"))
		sprintf(temp_buffer,"%3.2lf",cpu_load_5_minute[2]);
	else if(!strcmp(recv_buffer,"CPU2LOAD15"))
		sprintf(temp_buffer,"%3.2lf",cpu_load_15_minute[2]);

	else if(!strcmp(recv_buffer,"CPU3LOAD1"))
		sprintf(temp_buffer,"%3.2lf",cpu_load_1_minute[3]);
	else if(!strcmp(recv_buffer,"CPU3LOAD5"))
		sprintf(temp_buffer,"%3.2lf",cpu_load_5_minute[3]);
	else if(!strcmp(recv_buffer,"CPU3LOAD15"))
		sprintf(temp_buffer,"%3.2lf",cpu_load_15_minute[3]);

	
	else
		strcpy(temp_buffer,"-1");

	strncat(send_buffer,temp_buffer,send_length);

	return 0;
	}



// get total physical memory on the local machine
int get_total_physical_memory(int FAR *total_memory){
	MEMORYSTATUS memstatus;

	GlobalMemoryStatus(&memstatus);

	*total_memory=memstatus.dwTotalPhys;

	return 0;
	}


// get available physical memory on the local machine
int get_available_physical_memory(int FAR *avail_memory){
	MEMORYSTATUS memstatus;

	GlobalMemoryStatus(&memstatus);

	*avail_memory=memstatus.dwAvailPhys;

	return 0;
	}


// get total paged memory on the local machine
int get_total_paged_memory(int FAR *total_memory){
	MEMORYSTATUS memstatus;

	GlobalMemoryStatus(&memstatus);

	*total_memory=memstatus.dwTotalPageFile;

	return 0;
	}


// get available paged memory on the local machine
int get_available_paged_memory(int FAR *avail_memory){
	MEMORYSTATUS memstatus;

	GlobalMemoryStatus(&memstatus);

	*avail_memory=memstatus.dwAvailPageFile;

	return 0;
	}


// get total disk space, either on a local drive or on a share
int get_total_disk_space(char FAR *disk_name,ULARGE_INTEGER FAR *total_disk){
	int err;
	ULARGE_INTEGER temp_disk_caller;
	ULARGE_INTEGER temp_disk_space;
	ULARGE_INTEGER temp_disk_avail;

	printf("Finding total disk space on: '%s'\n",disk_name);

	err=GetDiskFreeSpaceEx(disk_name,&temp_disk_caller,&temp_disk_space,&temp_disk_avail);
	if(!err){
		printf("Error: GetDiskFreeSpaceEx() returned an error!\n");
		return 1;
		}
	
	// convert to KB
	total_disk->QuadPart=(temp_disk_space.QuadPart)/1024;

	return 0;
	}


// get total disk space, either on a local drive or on a share
int get_available_disk_space(char FAR *disk_name,ULARGE_INTEGER FAR *avail_disk){
	int err;
	ULARGE_INTEGER temp_disk_caller;
	ULARGE_INTEGER temp_disk_space;
	ULARGE_INTEGER temp_disk_avail;

	printf("Finding available disk space on: '%s'\n",disk_name);

	err=GetDiskFreeSpaceEx(disk_name,&temp_disk_caller,&temp_disk_space,&temp_disk_avail);
	if(!err){
		printf("Error: GetDiskFreeSpaceEx() returned an error!\n");
		return 1;
		}
	
	// convert to KB
	avail_disk->QuadPart=(temp_disk_avail.QuadPart)/1024;

	return 0;
	}




// EOF