/*******************************************************************
 *
 * PERF.C - Performance Data Functions
 * Copyright (c) 2000 Ethan Galstad (netsaint@linuxbox.com)
 * Last Modified: 02-12-2000
 * License: GPL
 *
 * Credits: Most of the perfdata code is based on the code found
 *          int CDC's BackOrifice 2000.  See http://www.bo2k.com
 *			for more information.
 *
 ******************************************************************/

#include <windows.h> 
#include <winperf.h> 
#include <stdlib.h> 
#include <stdio.h>
 
#include "include\perf.h"

 
#define INDEX_STR_LEN       10 
#define MACHINE_NAME_LEN    MAX_COMPUTERNAME_LENGTH+2 
#define MACHINE_NAME_SIZE   MACHINE_NAME_LEN+1 

LPTSTR      *gPerfTitleSz; 
LPTSTR      TitleData; 
 
// Globals 

TCHAR           INDEX_PROCTHRD_OBJ[2*INDEX_STR_LEN]; 
TCHAR           INDEX_COSTLY_OBJ[3*INDEX_STR_LEN]; 

TCHAR           gszMachineName[MACHINE_NAME_SIZE]; 
TCHAR           gszCurrentMachine[MACHINE_NAME_SIZE]; 

DWORD           gPerfDataSize = 50*1024;            // start with 50K 
PPERF_DATA      gpPerfData; 

HKEY            ghPerfKey = HKEY_PERFORMANCE_DATA;  // get perf data from this key 
HKEY            ghMachineKey = HKEY_LOCAL_MACHINE;  // get title index from this key 



ULARGE_INTEGER last_processor_load[4];
ULARGE_INTEGER this_processor_load[4];

DOUBLE cpu_load15_buffer[4][15];
DOUBLE cpu_load5_buffer[4][5];

int cpu_load15_buffer_index[4];
int cpu_load5_buffer_index[4];

extern DOUBLE cpu_load_1_minute[4];
extern DOUBLE cpu_load_5_minute[4];
extern DOUBLE cpu_load_15_minute[4];

extern DOUBLE average_cpu_load_1_minute;
extern DOUBLE average_cpu_load_5_minute;
extern DOUBLE average_cpu_load_15_minute;





/************************************************************************************************/
/************************************************************************************************/
/*																								*/
/*								PERF DATA TRAVERSAL FUNCTIONS									*/
/*																								*/
/************************************************************************************************/
/************************************************************************************************/



// FirstInstance() - Returns pointer to the first instance of pObject type. 
//                   If pObject is NULL then NULL is returned. 
 
PPERF_INSTANCE   FirstInstance (PPERF_OBJECT pObject){ 
    if (pObject) 
        return (PPERF_INSTANCE)((PCHAR) pObject + pObject->DefinitionLength); 
    else 
        return NULL; 
	}
 
 

// NextInstance() - Returns pointer to the next instance following pInst. 
//                  If pInst is the last instance, bogus data maybe returned. 
//                  The caller should do the checking. 
//                  If pInst is NULL, then NULL is returned. 

PPERF_INSTANCE   NextInstance (PPERF_INSTANCE pInst){ 
	PERF_COUNTER_BLOCK *pCounterBlock; 
	
    if (pInst){ 
        pCounterBlock = (PERF_COUNTER_BLOCK *)((PCHAR) pInst + pInst->ByteLength); 
        return (PPERF_INSTANCE)((PCHAR) pCounterBlock + pCounterBlock->ByteLength); 
		}
    else 
        return NULL; 
	}
 
 
 
 
// FindInstanceN() - Returns the Nth instance of pObject type.  
//                   If not found, NULL is returned.  0 <= N <= NumInstances. 
 
PPERF_INSTANCE FindInstanceN (PPERF_OBJECT pObject, DWORD N){ 
	PPERF_INSTANCE pInst; 
	DWORD          i = 0; 
	
    if (!pObject) 
        return NULL; 
    else if (N >= (DWORD)(pObject->NumInstances)) 
        return NULL; 
    else{ 

        pInst = FirstInstance (pObject); 
		
        while(i != N){ 
            pInst = NextInstance (pInst); 
            i++; 
			} 
		
        return pInst; 
		}
	}
 
 
 
 
// FindInstanceParent() - Returns the pointer to an instance that is the parent of pInst. 
//                        If pInst is NULL or the parent object is not found 
//                        then NULL is returned. 

PPERF_INSTANCE FindInstanceParent (PPERF_INSTANCE pInst, PPERF_DATA pData){ 
	PPERF_OBJECT    pObject; 
	
    if(!pInst) 
        return NULL; 
    else if (!(pObject = FindObject (pData, pInst->ParentObjectTitleIndex))) 
        return NULL; 
    else 
        return FindInstanceN (pObject, pInst->ParentObjectInstance); 
	} 


 
 
// InstanceName() - Returns the name of the pInst. 
//                  If pInst is NULL then NULL is returned. 

LPTSTR  InstanceName (PPERF_INSTANCE pInst){ 
    if (pInst) 
        return (LPTSTR) ((PCHAR) pInst + pInst->NameOffset); 
    else 
        return NULL; 
	}

 



// FirstObject() - Returns pointer to the first object in pData. 
//                 If pData is NULL then NULL is returned. 

PPERF_OBJECT FirstObject (PPERF_DATA pData){ 
    if (pData) 
        return ((PPERF_OBJECT) ((PBYTE) pData + pData->HeaderLength)); 
    else 
        return NULL; 
	}
 
 
 
// NextObject() - Returns pointer to the next object following pObject. 
//	              If pObject is the last object, bogus data maybe returned. 
//                The caller should do the checking. 
//                If pObject is NULL, then NULL is returned. 

PPERF_OBJECT NextObject (PPERF_OBJECT pObject){ 
    if (pObject) 
        return ((PPERF_OBJECT) ((PBYTE) pObject + pObject->TotalByteLength)); 
    else 
        return NULL; 
	}
 
 
 
 
// FindObject() - Returns pointer to object with TitleIndex.
//                If not found, NULL is returned. 

PPERF_OBJECT FindObject (PPERF_DATA pData, DWORD TitleIndex){	
	PPERF_OBJECT pObject; 
	DWORD i = 0; 
	
    if (pObject = FirstObject (pData)){
        while (i < pData->NumObjectTypes){ 
            if (pObject->ObjectNameTitleIndex == TitleIndex) 
                return pObject; 
			
            pObject = NextObject (pObject); 
            i++; 
			}
		}
		
	return NULL; 
	}
 


// FindObjectN() - Find the Nth object in pData.  
//                 If not found, NULL is returned. 0 <= N < NumObjectTypes. 

PPERF_OBJECT FindObjectN (PPERF_DATA pData, DWORD N){ 
	PPERF_OBJECT pObject; 
	DWORD        i = 0; 
	
    if (!pData) 
        return NULL; 
    else if (N >= pData->NumObjectTypes) 
        return NULL; 
    else{ 
        pObject = FirstObject (pData); 
		
        while (i != N){ 
            pObject = NextObject (pObject); 
            i++; 
			}
		
        return pObject; 
		}
	}




// FirstCounter() - Find the first counter in pObject. 
//                  Returns a pointer to the first counter.  If pObject is NULL 
//                  then NULL is returned. 

PPERF_COUNTER FirstCounter (PPERF_OBJECT pObject){ 
    if (pObject) 
        return (PPERF_COUNTER)((PCHAR) pObject + pObject->HeaderLength); 
    else 
        return NULL; 
	}
 
 
 
 
// NextCounter() - Find the next counter of pCounter. 
//                 If pCounter is the last counter of an object type, bogus data 
//                 maybe returned.  The caller should do the checking. 
//                 Returns a pointer to a counter.  If pCounter is NULL then 
//                 NULL is returned. 

PPERF_COUNTER NextCounter (PPERF_COUNTER pCounter){ 
    if (pCounter) 
        return (PPERF_COUNTER)((PCHAR) pCounter + pCounter->ByteLength); 
    else 
        return NULL; 
	}
 
 
 
 
// FindCounter() - Find a counter specified by TitleIndex. 
//                 Returns a pointer to the counter.  If counter is not found 
//                 then NULL is returned. 

PPERF_COUNTER FindCounter (PPERF_OBJECT pObject, DWORD TitleIndex){ 
	PPERF_COUNTER pCounter; 
	DWORD         i = 0; 
	
    if (pCounter = FirstCounter (pObject)) 
        while (i < pObject->NumCounters){ 
            if (pCounter->CounterNameTitleIndex == TitleIndex) 
                return pCounter; 
			
            pCounter = NextCounter (pCounter); 
            i++; 
			}
		
	return NULL; 
	}
 
 
 
 
//  CounterData() - Returns counter data for an object instance.  
//                  If pInst or pCount is NULL then NULL is returned.

PVOID CounterData (PPERF_INSTANCE pInst, PPERF_COUNTER pCount){ 
	PPERF_COUNTER_BLOCK pCounterBlock; 
	
    if (pCount && pInst) 
	{ 
        pCounterBlock = (PPERF_COUNTER_BLOCK)((PCHAR)pInst + pInst->ByteLength); 
        return (PVOID)((PCHAR)pCounterBlock + pCount->CounterOffset); 
	} 
    else 
        return NULL; 
	}




//  SetPerfIndexes() - Setup the perf data indexes. 

int SetPerfIndexes(void){ 
	LPTSTR  TitleBuffer; 
	LPTSTR  *Title; 
	DWORD   Last; 	
	
	if(GetPerfTitleSz (ghMachineKey, ghPerfKey, &TitleBuffer, &Title, &Last)!=ERROR_SUCCESS) return -1;	

	PX_PROCESSOR                       = GetTitleIdx (Title, Last, PN_PROCESSOR); 
	PX_PROCESSOR_CPU                   = GetTitleIdx (Title, Last, PN_PROCESSOR_CPU); 
	PX_PROCESSOR_PRIV                  = GetTitleIdx (Title, Last, PN_PROCESSOR_PRIV); 
	PX_PROCESSOR_USER                  = GetTitleIdx (Title, Last, PN_PROCESSOR_USER); 


	LocalFree (TitleBuffer); 
	LocalFree (Title); 
	
	return 0;
	}




//  GetPerfTitleSz() - Retrieves the performance data title strings. 
// 
//   This call retrieves english version of the title strings. 
// 
//   For NT 3.1, the counter names are stored in the "Counters" value 
//   in the ...\perflib\009 key.  For 3.5 and later, the 009 key is no 
//      longer used.  The counter names should be retrieved from "Counter 009" 
//      value of HKEY_PERFORMANCE_KEY. 
// 
//      Caller should provide two pointers, one for buffering the title 
//      strings the other for indexing the title strings.  This function will 
//      allocate memory for the TitleBuffer and TitleSz.  To get the title 
//      string for a particular title index one would just index the TitleSz. 
//      *TitleLastIdx returns the highest index can be used.  If TitleSz[N] is 
//      NULL then there is no Title for index N. 
// 
//      Example:  TitleSz[20] points to titile string for title index 20. 
// 
//      When done with the TitleSz, caller should LocalFree(*TitleBuffer). 
// 
//      This function returns ERROR_SUCCESS if no error. 

DWORD   GetPerfTitleSz (HKEY hKeyMachine, HKEY hKeyPerf, LPTSTR *TitleBuffer, LPTSTR  *TitleSz[], DWORD *TitleLastIdx){ 
	HKEY  hKey1; 
	HKEY    hKey2; 
	DWORD   Type; 
	DWORD   DataSize; 
	DWORD   dwR; 
	DWORD   Len; 
	DWORD   Index; 
	LPTSTR  szCounterValueName; 
	LPTSTR  szTitle; 
	
    // Initialize 
    // 
    hKey1        = NULL; 
    hKey2        = NULL; 
    *TitleBuffer = NULL; 
    *TitleSz     = NULL; 
	
    // Open the perflib key to find out the last counter's index and system version. 
    
	dwR = RegOpenKeyEx (hKeyMachine, 
		TEXT("software\\microsoft\\windows nt\\currentversion\\perflib"), 
		0, 
		KEY_READ, 
		&hKey1); 
    if (dwR != ERROR_SUCCESS) 
        goto done; 
	
    // Get the last counter's index so we know how much memory to allocate for TitleSz 
    
	DataSize = sizeof (DWORD); 
    dwR = RegQueryValueEx (hKey1, TEXT("Last Counter"), 0, &Type, (LPBYTE)TitleLastIdx, &DataSize); 
    if (dwR != ERROR_SUCCESS) 
        goto done; 
	
    // NT 1.0a or later.  Get the counters in key HKEY_PERFORMANCE_KEY 
    //  and from value "Counter 009" 
    
	szCounterValueName = TEXT("Counter 009"); 
    hKey2 = hKeyPerf; 

	
    // Find out the size of the data. 
    // 
    dwR = RegQueryValueEx (hKey2, szCounterValueName, 0, &Type, 0, &DataSize); 
    if (dwR != ERROR_SUCCESS) 
        goto done; 

	
    // Allocate memory 
    // 
    *TitleBuffer = (LPTSTR)LocalAlloc (LMEM_FIXED, DataSize); 
    if (!*TitleBuffer){ 
        dwR = ERROR_NOT_ENOUGH_MEMORY; 
        goto done; 
		}
	
    *TitleSz = (LPTSTR *)LocalAlloc (LPTR, (*TitleLastIdx+1) * sizeof (LPTSTR)); 
    if (!*TitleSz){ 
        dwR = ERROR_NOT_ENOUGH_MEMORY; 
        goto done; 
		}
	
	
	// Query the data 
 	dwR = RegQueryValueEx (hKey2, szCounterValueName, 0, &Type, (BYTE *)*TitleBuffer, &DataSize); 
    if (dwR != ERROR_SUCCESS) 
        goto done; 
	
	
	// Setup the TitleSz array of pointers to point to beginning of each title string. 
    // TitleBuffer is type REG_MULTI_SZ. 
    
	szTitle = *TitleBuffer; 
	
    while (Len = lstrlen (szTitle)){ 
        Index = atoi (szTitle); 
		
        szTitle = szTitle + Len +1; 
		
        if (Index <= *TitleLastIdx) 
            (*TitleSz)[Index] = szTitle; 
		
        szTitle = szTitle + lstrlen (szTitle) +1; 
		}
	
	
	
done: 
	
    // Done. Now cleanup! 
    
	if (dwR != ERROR_SUCCESS){ 
        // There was an error, free the allocated memory 
        // 
        if (*TitleBuffer) LocalFree (*TitleBuffer); 
        if (*TitleSz)     LocalFree (*TitleSz); 
		}
	
    // Close the hKeys. 

    if (hKey1) RegCloseKey (hKey1); 
    if (hKey2 && hKey2 != hKeyPerf) RegCloseKey (hKey2); 

	
    return dwR; 
	}

 



//  GetTitleIdx() - Searches Titles[] for Name.  Returns the index found. 

DWORD GetTitleIdx(LPTSTR Title[], DWORD LastIndex, LPTSTR Name){ 
	DWORD   Index; 
	
	for (Index = 0; Index <= LastIndex; Index++) 
		if (Title[Index]) 
			if (!lstrcmpi (Title[Index], Name)) 
				return Index; 
			
	return 0; 
	}





//  SetLocalMachine() - Set local machine as performance data focus. 
//                      Sets ghPerfKey, ghMachineKey, gszMachineName, gszCurrentMachine 

void PerfData_SetLocalMachine(void){ 
	TCHAR   szName[MACHINE_NAME_SIZE]; 
	DWORD   dwSize = MACHINE_NAME_SIZE; 
	
	// close remote connections, if any 
	if (ghPerfKey!=HKEY_PERFORMANCE_DATA) RegCloseKey(ghPerfKey); 
	if (ghMachineKey!=HKEY_LOCAL_MACHINE) RegCloseKey(ghMachineKey); 
		
	// set to registry keys on local machine 
	ghPerfKey    = HKEY_PERFORMANCE_DATA; 
	ghMachineKey = HKEY_LOCAL_MACHINE; 
	
	// get computer name 
	GetComputerName (szName, &dwSize); 
	
	if (szName[0] != '\\' || szName[1] != '\\'){ 
	    // must have two '\\' 
		wsprintf (gszMachineName, TEXT("\\\\%s"), szName); 
		lstrcpy (gszCurrentMachine, gszMachineName); 
		} 
	else{ 
		lstrcpy (gszMachineName, szName); 
		lstrcpy (gszCurrentMachine, gszMachineName); 
		} 
	
	} 




//  ConnectComputer() - Connect to a computer with name entered in PVIEW_COMPUTER. 
//                      If a new connection is made, then return TRUE else return FALSE. 
//                      Sets gszCurrentMachine, ghPerfKey, and ghMachineKey

int PerfData_ConnectComputer(char *svName){ 
	HKEY    hKey; 
	TCHAR   szTemp[MACHINE_NAME_SIZE]; 
	BOOL    bResult = TRUE; 
	
	if(svName==NULL){
		// Connect to local machine
		PerfData_SetLocalMachine(); 
		SetPerfIndexes(); 
	
		return 0;
		}
	
	// Connect to remote machine 
	
	lstrcpyn(szTemp,svName,MACHINE_NAME_SIZE);
	
	if(RegConnectRegistry(szTemp, HKEY_PERFORMANCE_DATA, &hKey)!=ERROR_SUCCESS) return -1;
	
	lstrcpy(gszCurrentMachine, szTemp); 

	if(ghPerfKey!=HKEY_PERFORMANCE_DATA) RegCloseKey (ghPerfKey); 	
	ghPerfKey = hKey; 
		
	// we also need to get the remote machine's title indexes. 
	
	if (ghMachineKey != HKEY_LOCAL_MACHINE) RegCloseKey (ghMachineKey); 
			
	if (RegConnectRegistry (gszCurrentMachine, HKEY_LOCAL_MACHINE, &hKey) == ERROR_SUCCESS) 
		ghMachineKey = hKey; 
	else 
		ghMachineKey = HKEY_LOCAL_MACHINE; 

	SetPerfIndexes(); 
	return 0;
	}



// RefreshPerfData() - Get a new set of performance data. pData should be NULL initially. 
// GetPerfData() - Get a new set of performance data. 
//      *ppData should be NULL initially. 
//      This function will allocate a buffer big enough to hold the 
//      data requested by szObjectIndex. 
// 
//      *pDataSize specifies the initial buffer size.  If the size is 
//      too small, the function will increase it until it is big enough 
//      then return the size through *pDataSize.  Caller should 
//      deallocate *ppData if it is no longer being used. 
// 
//      Returns ERROR_SUCCESS if no error occurs. 
// 
//      Note: the trial and error loop is quite different from the normal 
//            registry operation.  Normally if the buffer is too small, 
//            RegQueryValueEx returns the required size.  In this case, 
//            the perflib, since the data is dynamic, a buffer big enough 
//            for the moment may not be enough for the next. Therefor, 
//            the required size is not returned. 
// 
//            One should start with a resonable size to avoid the overhead 
//            of reallocation of memory. 

DWORD GetPerfData (HKEY hPerfKey, LPTSTR szObjectIndex, PPERF_DATA *ppData, DWORD *pDataSize){ 	
	DWORD   DataSize; 
	DWORD   dwR; 
	DWORD   Type; 
	
    if (!*ppData) 
        *ppData = (PPERF_DATA) LocalAlloc (LMEM_FIXED, *pDataSize); 	
	
    do{ 
        DataSize = *pDataSize; 
        dwR = RegQueryValueEx (hPerfKey, 
			szObjectIndex, 
			NULL, 
			&Type, 
			(BYTE *)*ppData, 
			&DataSize); 
		
        if(dwR == ERROR_MORE_DATA){ 
            LocalFree (*ppData); 
            *pDataSize += 1024; 
            *ppData = (PPERF_DATA) LocalAlloc (LMEM_FIXED, *pDataSize); 
			}
		
        if(!*ppData){ 
            LocalFree (*ppData); 
            return ERROR_NOT_ENOUGH_MEMORY; 
			} 
		
		}while(dwR == ERROR_MORE_DATA); 
	
    return dwR; 
	} 
 
  
 
 

PPERF_DATA RefreshPerfData (HKEY hPerfKey, LPTSTR szObjectIndex, PPERF_DATA pData, DWORD *pDataSize){ 

    if (GetPerfData (hPerfKey, szObjectIndex, &pData, pDataSize) == ERROR_SUCCESS) 
        return pData; 
    else 
        return NULL; 
	} 
 







/************************************************************************************************/
/************************************************************************************************/
/*																								*/
/*									PROCESSOR LOAD FUNCTIONS									*/
/*																								*/
/************************************************************************************************/
/************************************************************************************************/



// calculates processor load (runs as a separate thread)
void PerfData_GetProcessorLoad(void){	
	PPERF_OBJECT pObject;					// pointer to an object
	PPERF_COUNTER pCounterID; 				// pointer to a counter
	PPERF_INSTANCE pInstance;				// pointer to an instance
	//int InstanceIndex;
	int current_processor;
	int x;

	// initialize processor load variables
	for(current_processor=0;current_processor<4;current_processor++){
		for(x=0;x<5;x++)
			cpu_load5_buffer[current_processor][x]=0.0;
		for(x=0;x<15;x++)
			cpu_load15_buffer[current_processor][x]=0.0;
		cpu_load_1_minute[current_processor]=0.0;
		cpu_load_5_minute[current_processor]=0.0;
		cpu_load_15_minute[current_processor]=0.0;
		cpu_load5_buffer_index[current_processor]=0;
		cpu_load15_buffer_index[current_processor]=0;
		}

	// set the local machine as the focus of the performance data
	PerfData_ConnectComputer(NULL);

	// get performance data 
	gpPerfData=RefreshPerfData(ghPerfKey, INDEX_PROCTHRD_OBJ, gpPerfData, &gPerfDataSize); 
	if(gpPerfData==NULL)
		printf("gdPerfData is NULL!\n");

	// Let's get information about processor
	pObject=FindObject(gpPerfData,PX_PROCESSOR); 
	if(pObject==NULL)
		printf("pObject is NULL!\n");

	// Get initial CPU utilization on up to first 4 processors
	pCounterID=FindCounter(pObject,PX_PROCESSOR_CPU);
	for(current_processor=0,pInstance=FirstInstance(pObject);current_processor<4 && pInstance!=NULL && current_processor<(DWORD)(pObject->NumInstances);current_processor++,pInstance=NextInstance(pInstance))
		this_processor_load[current_processor]=*(ULARGE_INTEGER *)CounterData(pInstance,pCounterID);


	while(1){
	
		// sleep 60 seconds
		Sleep(60000);

		printf("Refreshing perf data...\n");

		// get performance data 
		gpPerfData=RefreshPerfData(ghPerfKey, INDEX_PROCTHRD_OBJ, gpPerfData, &gPerfDataSize); 

		// Let's get information about processor
		pObject=FindObject(gpPerfData,PX_PROCESSOR); 

		// Get CPU utilization on up to first 4 processors
		pCounterID=FindCounter(pObject,PX_PROCESSOR_CPU);
		for(current_processor=0,pInstance=FirstInstance(pObject);current_processor<4 && pInstance!=NULL && current_processor<(DWORD)(pObject->NumInstances);current_processor++,pInstance=NextInstance(pInstance)){
			last_processor_load[current_processor]=this_processor_load[current_processor];
			this_processor_load[current_processor]=*(ULARGE_INTEGER *)CounterData(pInstance,pCounterID);
			calculate_processor_load(current_processor);
			}

		// calculate average utilization for up to first 4 processors
		calculate_average_processor_load(current_processor);
		}
	
	return;
	}



// calculates 1, 5, and 15 minutes load averages for a particular CPU
void calculate_processor_load(int processor_number){
	ULARGE_INTEGER cpu_difference;
	DWORD cpu_load_inv=0;
	DOUBLE cpu_load_inv_frac=0.0;
	DOUBLE cpu_load_frac=0.0;
	int index=0;
	DOUBLE running_total=0.0;

	cpu_difference.QuadPart=this_processor_load[processor_number].QuadPart-last_processor_load[processor_number].QuadPart;
	cpu_load_inv=cpu_difference.QuadPart/600000;
	cpu_load_inv_frac=(cpu_load_inv/1000.0);
	cpu_load_frac=1.0-cpu_load_inv_frac;

	// calculate 1 minute average CPU load
	cpu_load_1_minute[processor_number]=(cpu_load_frac*100.0);

	// store data for 5 minute average calculation
	if(cpu_load5_buffer_index[processor_number]>=4)
		cpu_load5_buffer_index[processor_number]=1;
	else
		cpu_load5_buffer_index[processor_number]++;
	cpu_load5_buffer[processor_number][cpu_load5_buffer_index[processor_number]-1]=cpu_load_1_minute[processor_number];

	// calculate 5 minute average CPU load
	for(index=0,running_total=0.0;index<5;index++)
		running_total+=cpu_load5_buffer[processor_number][index];
	cpu_load_5_minute[processor_number]=running_total/5.0;

	// store data for 15 minute average calculation
	if(cpu_load15_buffer_index[processor_number]>=14)
		cpu_load15_buffer_index[processor_number]=1;
	else
		cpu_load15_buffer_index[processor_number]++;
	cpu_load15_buffer[processor_number][cpu_load15_buffer_index[processor_number]-1]=cpu_load_1_minute[processor_number];

	// calculate 15 minute average CPU load
	for(index=0,running_total=0.0;index<15;index++)
		running_total+=cpu_load15_buffer[processor_number][index];
	cpu_load_15_minute[processor_number]=running_total/15.0;


	printf("CPU %d LOAD: [1-min]=%3.2lf%%, [5-min]=%3.2lf%%, [15-min]=%3.2lf%%\n",processor_number,cpu_load_1_minute[processor_number],cpu_load_5_minute[processor_number],cpu_load_15_minute[processor_number]);

	return;
	}



// calculates average processor load across (up to) first four CPUs
void calculate_average_processor_load(total_processors){
	int current_processor;
	DOUBLE temp_load_1_minute=0.0;
	DOUBLE temp_load_5_minute=0.0;
	DOUBLE temp_load_15_minute=0.0;

	for(current_processor=0;current_processor<total_processors;current_processor++){
		temp_load_1_minute+=cpu_load_1_minute[current_processor];
		temp_load_5_minute+=cpu_load_5_minute[current_processor];
		temp_load_15_minute+=cpu_load_15_minute[current_processor];
		}

	average_cpu_load_1_minute=temp_load_1_minute/total_processors;
	average_cpu_load_5_minute=temp_load_5_minute/total_processors;
	average_cpu_load_15_minute=temp_load_15_minute/total_processors;
	}